/*
 * $Id: render.c,v 1.1.1.1 2001/12/19 14:45:33 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */

/**
 * Render 3D.
 *
 * @version $Revision: 1.1.1.1 $
 */
#include <string.h>

#include "render.h"
#include "vector.h"
#include "re.h"

static int min[SCREEN_HEIGHT], max[SCREEN_HEIGHT];

static void scanEdge(Vector *v1, Vector *v2) {
  int l, addx, addy, x, y, i;

  l = absN((int)((v2->y>>8) - (v1->y>>8))) + 1;
	
  addx = (v2->x - v1->x) / l;
  addy = (v2->y - v1->y) / l;

  x = v1->x;
  y = v1->y;

  for ( i=0 ; i<l ; i++, x+=addx, y+=addy ) {
    int px, py;
    py = y>>8; px = x>>8;
    if ( py < 0 || py >= SCREEN_HEIGHT ) continue;
    if ( min[py] > px ){
      min [py] = px;
    }
    if ( max[py] < px ){
      max [py] = px;
    }
  }
}

#define EDGE_COLOR 2
#define POLYGON_COLOR 1

static void drawPolygon(Vector *v1, Vector *v2, Vector *v3, Vector *v4) {
  int top, btm, y;
  int sx, lgt;
  unsigned char *vbOfs;

  top = +2147483647;
  btm = -2147483647;
  if ( top > (int)v1->y ) top = (int)v1->y;
  if ( top > (int)v2->y ) top = (int)v2->y;
  if ( top > (int)v3->y ) top = (int)v3->y;
  if ( btm < (int)v1->y ) btm = (int)v1->y;
  if ( btm < (int)v2->y ) btm = (int)v2->y;
  if ( btm < (int)v3->y ) btm = (int)v3->y;
  if ( v4 != NULL ) {
    if ( top > (int)v4->y ) top = (int)v4->y;
    if ( btm < (int)v4->y ) btm = (int)v4->y;
  }
  top >>= 8;
  btm >>= 8;
  if ( top < 0 ) top = 0;
  if ( btm > SCREEN_HEIGHT ) btm = SCREEN_HEIGHT;
				
  for ( y=top ; y<btm ; y++ ) {
    min[y] = +2147483647;
    max[y] = -2147483647;
  }

  scanEdge(v1, v2);
  scanEdge(v2, v3);
  if ( v4 != NULL ) {
    scanEdge(v3, v4);
    scanEdge(v4, v1);
  } else {
    scanEdge(v3, v1);
  }
		
  for ( y=top ; y<btm ; y++ ){
    if ( min[y] == +2147483647 ) continue;
    sx = min[y];
    if ( sx < 0 ) sx = 0;
    lgt = max[y];
    if ( lgt >= SCREEN_WIDTH ) lgt = SCREEN_WIDTH-1;
    lgt -= sx;
    if ( lgt <= 0 ) continue;
    //vbOfs = &(vbuf[y*SCREEN_WIDTH+sx]);
    vbOfs = &(vbuf[(y<<7)+sx]);
    if ( y==top || y==btm-1 ) {
      memset(vbOfs, EDGE_COLOR, lgt);
    } else {
      *vbOfs = EDGE_COLOR;
      if ( lgt > 2 ) {
	memset(vbOfs+1, POLYGON_COLOR, lgt-2);
      }
      *(vbOfs+lgt) = EDGE_COLOR;
    }
  }
}

#define TVTX_MAX 64

static Vector tvtx[TVTX_MAX];

void drawPrimitive(Vector *vertices, int vn, int in, int ofsX, int ofsY, int ofsZ) {
  int i;
  int z;

  for ( i=0 ; i<vn ; i++ ) {
    z = vertices[i].z + ofsZ;
    tvtx[i].x = ((vertices[i].x+ofsX) << 14) / z + (SCREEN_WIDTH<<7);
    tvtx[i].y = ((vertices[i].y+ofsY) << 14) / z + (SCREEN_HEIGHT<<5);
  }

  switch ( in ) {
  case 3:
    for ( i=0 ; i<vn ; i+=3 ) {
      drawPolygon(&(tvtx[i]), &(tvtx[i+1]), &(tvtx[i+2]), NULL);
    }
    break;
  case 4:
    for ( i=0 ; i<vn ; i+=4 ) {
      drawPolygon(&(tvtx[i]), &(tvtx[i+1]), &(tvtx[i+2]), &(tvtx[i+3]));
    }
    break;
  }
}
